/* N-body calculation using Greenspan 1st-order discrete mechanics.	*/
/* Uses 3*N*N (of 4*N*N) processors, communication by "left subscripts".*/
/* Incorporates xyz-parallelism, diagonal processors store particles.	*/
/* File:  nb3dg2.cs							*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

/* Max number of bodies in the calculation.  N = 64 uses 16k vprocs.	*/

#define N 64

int  NB;     /* Number of bodies in system, <= N	*/
int  PX;     /* Controls how much of state is printed   */
float G;     /* Gravitational constant			*/

#define TWOPI 6.2831853

/* Particles, and pairs of particles.  Diagonal entries have particle	*/
/* states.  Some variables could be overloaded. 			*/

shape [4][N][N]pair;

/* x0, x1 - Current, new positions;      v0, v1 - velocities	*/
/* d0 - diplacements from x0's;	    r0 = || d0||		*/
/* m - mass;      mj - scaled mass;      ac - acceleration 	*/
/* w - working							*/

float:pair x0, x1, v0, v1, m, mj, ac, d0, r0, ax, w;

/* i - row number;    j - column number;    c - coordinate number	*/
/* r - index of "right" neighbor along c axis				*/

int:pair i, j, c, r;

/* here - processor represents actual particle		*/
/* act  - processor is active pair			*/

bool:pair here, act;

/* Pairs are represented symmetrically.	 Setup initializes these.	*/

void setup()
 { with (pair)
    { bool:pair sm;
      /* Initialize coordinates */
      c = pcoord(0);  i = pcoord(1);  j = pcoord(2);
      r = (c < 2) ? (c+1) : 0;
      /* Identify active processors */
      sm = (i < NB) & (j < NB) & c < 3;
      act = sm & (i != j);  here = sm & (i == j);
    }
 }

FILE *datap;      /* Stream for reading data */

/* Load inital positions, velocities, also masses, gravitational const. */

void init()
 { int k;  float a, b, c;
   with (pair)
   {for ( k = 0; k < NB; k++ )
     { fscanf(datap, "%e%e%e", &a, &b, &c);
       [0][k][k]x0 = a;  [1][k][k]x0 = b;  [2][k][k]x0 = c;
     };
    for ( k = 0; k < NB; k++ )
     { fscanf(datap, "%e%e%e", &a, &b, &c);
       [0][k][k]v0 = a;  [1][k][k]v0 = b;  [2][k][k]v0 = c;
     };
    for ( k = 0; k < NB; k++ )
      { fscanf(datap, "%e", &a);
        [0][k][k]m = [1][k][k]m = [2][k][k]m = a;
      };
    fscanf(datap, "%e", &G);
   }
 }

/* Insert nx asteroids at radius ar with total mass am.  This assumes	*/
/* that an orbit of radius 1 about origin corresponds to velocity 2*pi	*/
/* (e.g., au-earth-yr units).  Uses parallel trig.			*/

void mkast(int nx, float ar, float am)
 { float:pair sx, cx, arg;  bool:pair sm;
   float spd;  int nn;
   nn = NB + nx;
   spd = TWOPI/sqrt(ar);
   with (pair) 
   { sm = (i < nn) & (j < nn) & c < 3;
     act = sm & (i != j);  here = sm & (i == j);
     where ( (NB <= i) & (i < nn) & here )
      { arg = (TWOPI*(i - NB))/nx;
        sx = sin(arg);  cx = cos(arg);
        where ( c == 0 ) { x0 =  cx;  v0 = -sx; }
        where ( c == 1 ) { x0 =  sx;  v0 =  cx; }
        where ( c == 2 ) { x0 = 0.0;  v0 = 0.0; }
        x0 *= ar;  v0 *= spd;  m = am/nx;
      }}  /* end where, end with (pair) */
   NB = nn;
 }


/* Print current state of system according to PX.  See comments for	*/
/* main program for PX settings.					*/

void print_state()
 { int k, np;
   if ( (PX >= 0) && (PX < NB) ) np = PX; else np = NB;
   printf(
  "\n       x         y         z         vx        vy        vz        m\n");
   for ( k = 0; k < np; k++ )
     printf("%2d %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %9.2e\n",
            k, [0][k][k]x0, [1][k][k]x0, [2][k][k]x0,
               [0][k][k]v0, [1][k][k]v0, [2][k][k]v0, [0][k][k]m);
 }

/* Print x-components of accelerations in nx by nx processors (for testing). */

/* void print_xac(int nx)
 { int u, v;
   for ( u = 0; u < nx; u++)
    { for ( v = 0;  v < nx; v++)
        printf("9.2e ", [0][u][v]ac);
      printf("\n");
    }
 } */

/* Calculate accelerations.  r0, d0 already give displacement, distance  */
/* for current position.						 */

void accel()
 { float:pair d1, r1;
   ac = 0.0;  /* Everywhere */
   where (act)
    { d1 = [c][j][j]x1;  x1 = [c][i][i]x1;
      d1 = d1 - x1;		/* New displacement component */
      r1 = d1*d1;  [r][i][j]ac = r1;	/* For sum of squares */
      ac += r1;  [r][i][j]w = ac;  r1 = sqrt(w + r1);
      ac = mj*(d0+d1)/(r0*r1*(r0+r1));
      [c][i][i]ac += ac;        /* Calculate sum of accelerations */
    } /* end where (act) */
  }

/* Calculate initial accelerations */

void accel1()
 { ac = 0.0;  /* Everywhere */
   where (act)
    { d0 = [c][j][j]x0;  x0 = [c][i][i]x0;
      d0 = d0 - x0;		/* New displacement component */
      r0 = d0*d0;  [r][i][j]ac = r0;	/* For sum of squares */
      ac += r0;  [r][i][j]w = ac;  w += r0; r0 = sqrt(w);
      ac = mj*d0/(r0*w);
      [c][i][i]ac += ac;        /* Calculate sum of accelerations */
    } /* end where (act) */
 }

/* Greenspan's method.  Performs ns steps of time dt			*/
/* with solution tolerance eps.  Returns number of evaluations.		*/
/* Leaves number of iterations which were stopped at limit in cfail.	*/
/* Some printing for debugging is provided in comments.			*/

int cfail;   /* Counts number of iterations that failed to converge */

int green1(int ns, float dt, float eps)
 { int k, ic, v, lim=10;  float hdt, mx, my;
   v = 0;  hdt = 0.5*dt;
   with (pair)
   { float:pair nx;
     for (k = 0; k < ns; k++)
       { ic = 0;  accel1();
	 /* This is a do - while - repeat loop: */
   	 do
          { v++;  ic++;
	    where (here)
	    { nx = v0 + dt*ac;		/* New velocity	*/
	      mx = (>?= fabs(nx - v1));	/* Max change	*/
	      v1 = nx;
	      nx = x0 + hdt*(v0 + v1);	/* New position	*/
	      my = (>?= fabs(nx - x1));	/* Max change	*/
	      if (my > mx) mx = my;	/* Max change x, v */
	      x1 = nx; }
	    if ( (mx < eps) || ( ic == lim ) ) break;
	    accel(); }
	 while (1);
         where (here) { x0 = x1;  v0 = v1; };  if (ic == lim) cfail++;
       } ;   /* end for */
    }  /* end  with (pair) */
   return(v);
 }

/* Main program:  reads problem and control information from data file	*/
/* named by user and performs calculation.				*/
/* Print control:							*/
/*	  0 - just performance data					*/
/*	< 0 - all							*/
/*	> 0 - number given						*/

void main()
 { int ns, np, nv, k, vl;  float dt, eps, ar, am;
   char fnm[50];
   /* Performs np integrations, each ns steps of dt */
   printf("Data file> ");  scanf("%s", fnm);
   printf("Print ctl> ");  scanf("%d", &PX);
   datap = fopen(fnm, "r");
   fscanf(datap, "%d", &NB);
   setup();   /* Setup geometric structure */
   init();    /* Read data for individual bodies */
   /* Read "asteroid" descriptions */
   while ( fscanf(datap, "%d", &ns), ns > 0)
    { fscanf(datap, "%e%e", &ar, &am);  mkast(ns, ar, am); }
   fscanf(datap, "%d%d%e%e", &ns, &np, &dt, &eps);
   fclose(datap);
   printf("\n   G = %13.6e  delta-t = %12.5e  epsilon = %12.5e\n",
                 G, dt, eps);
   printf("\n simulated time = 0");
   print_state();
   /* Spread masses and scale */
   with (pair) where (act) { m = [c][j][j]m;  mj = G*m; }
   for ( k = 0; k < np; k++ )
    { CM_timer_clear(1);  CM_timer_start(1);
      cfail = 0;
      vl = green1(ns, dt, eps);
      CM_timer_stop(1);  CM_timer_print(1);
      printf(
        "\n simulated time = %9.3f  %d evaluations  %d iterations limited\n",
	             (k + 1)*ns*dt, vl,            cfail);
      print_state();
    };
 }

/*********************************************************************
  Data files:

    N			No. of explicit bodies

    x y z		Position of body 0
      :                      :
    x y z		Position of body N-1

    vx vy va		Velocity of body 0
       :                     :
    vx vy va		Velocity of body N-1

    m			Mass of body 0
    :			     :
    m			Mass of body N-1

    G			Gravitational constant

    nx ar am		Number, radius, mass of asteroid group
       :                        :     0 or more groups
    nx ar am		Number, radius, mass of asteroid group
    0			Stopper

    ns			Number of time steps per phase
    np			Number of phases
    dt			delta-t (per time step)
    eps			epsilon (solution tolerance)

*************************************************************************/
